#include "../src/baseSimulator.h"
#include "../src/baseSimulator.hh"

#include "gtest/gtest.h"

#include "1D/Architecture.test.cpp"
#include "1D/BigDroplet.test.cpp"
#include "1D/Continuous.test.cpp"
#include "hybrid/Hybrid.test.cpp"